#include "game.h"

Game::Game(abfw::Platform* platform) :
	abfw::Application(platform),
	sprite_renderer_(NULL),
	controller_manager_(SCE_CTRL_MODE_DIGITALANALOG)
	
{
}

void Game::Init()
{
	// load the font to draw the on-screen text
	bool font_loaded = font_.Load("comic_sans", *platform_);
	if(!font_loaded)
	{
		std::cout << "Font failed to load." << std::endl;
		exit(-1);
	}

	// create the main room
	MyMap.ChangeRoom();

	// initialise position of player and gun
	MyPlayer.MoveTo(MyMap.position().x, MyMap.position().y, PlayerZCoordinate);
	PlayerWeapon.MoveTo(MyMap.position().x + GunOffset, MyMap.position().y, FrontOfPlayer);
	
	// set the dimensions of the camera
	MyCamera.set_width(platform_->width());
	MyCamera.set_height(platform_->height());

	// initialise the frame count to 0
	frame_count_ = 0;

	// set the game state to the start screen
	current_game_screen_ = StartScreen;

	// create a sprite renderer to draw the sprites
	sprite_renderer_ = platform_->CreateSpriteRenderer();	
}

void Game::CleanUp()
{
	// free up the sprite renderer
	delete sprite_renderer_;
	sprite_renderer_ = NULL;
}

bool Game::Update(float ticks)
{
	// calculate the frame rate
	frame_rate_ = 1.0f / ticks;
	
	// update the controller manager
	controller_manager_.Update();

	// run all the game logic when on the game screen
	if(current_game_screen_ == GameScreen)
	{
		// update the player
		MyPlayer.Update(controller_manager_, MyMap.Platforms);

		// centre the camera on the player's position
		MyCamera.MoveTo(MyPlayer.position().x, MyPlayer.position().y, 0);

		// keep the camera within the boundaries of the map
		if(MyCamera.position().x - MyCamera.HalfSize(MyCamera.width()) < MyMap.position().x - MyMap.HalfSize(MyMap.width()))
			MyCamera.SetX(MyMap.position().x - MyMap.HalfSize(MyMap.width()) + MyCamera.HalfSize(MyCamera.width()));

		else if(MyCamera.position().x + MyCamera.HalfSize(MyCamera.width()) > MyMap.position().x + MyMap.HalfSize(MyMap.width()))
			MyCamera.SetX(MyMap.position().x + MyMap.HalfSize(MyMap.width()) - MyCamera.HalfSize(MyCamera.width()));
	
		if(MyCamera.position().y - MyCamera.HalfSize(MyCamera.height()) < MyMap.position().y - MyMap.HalfSize(MyMap.height()))
			MyCamera.SetY(MyMap.position().y - MyMap.HalfSize(MyMap.height()) + MyCamera.HalfSize(MyCamera.height()));

		else if(MyCamera.position().y + MyCamera.HalfSize(MyCamera.height()) > MyMap.position().y + MyMap.HalfSize(MyMap.height()))
			MyCamera.SetY(MyMap.position().y + MyMap.HalfSize(MyMap.height()) - MyCamera.HalfSize(MyCamera.height()));

		// keep the player within the boundaries of the map
		if(MyPlayer.position().x - MyPlayer.HalfSize(MyPlayer.width()) < RoomLeft)
			MyPlayer.SetX(RoomLeft + MyPlayer.HalfSize(MyPlayer.width()));

		else if(MyPlayer.position().x + MyPlayer.HalfSize(MyPlayer.width()) > RoomRight)
			MyPlayer.SetX(RoomRight - MyPlayer.HalfSize(MyPlayer.width()));

		if(MyPlayer.position().y - MyPlayer.HalfSize(MyPlayer.height()) < RoomTop)
			MyPlayer.SetY(RoomTop + MyPlayer.HalfSize(MyPlayer.height()));

		else if(MyPlayer.position().y + MyPlayer.HalfSize(MyPlayer.height()) > RoomBottom)
			MyPlayer.SetY(RoomBottom - MyPlayer.HalfSize(MyPlayer.height()));

		// aim the player's gun in the direction of the right stick
		PlayerWeapon.Aim(controller_manager_.right_stick_x_axis(), controller_manager_.right_stick_y_axis());

		// equip the gun upgrade if the player has it
		if(MyPlayer.has_gun_upgrade() == true)
			PlayerWeapon.set_fire_rate(ImprovedFireRate);

		// create a new bullet when the player fires the gun
		if((controller_manager_.right_stick_x_axis() != 0 || controller_manager_.right_stick_y_axis() != 0) && PlayerWeapon.can_fire() == true)
		{
			if((frame_count_ % PlayerWeapon.fire_rate()) == 0)
			{
				PlayerWeapon.CreateProjectile(PlayerWeapon.position().x, PlayerWeapon.position().y, PlayerWeapon.orientation());
			}
		}

		// when the right shoulder button is held down, the player can melee attack
		if(controller_manager_.buttons_down() & SCE_CTRL_R && MyPlayer.has_stick() == true && MyPlayer.has_sword() == false)
			PlayerWeapon.Melee(MyMap.Enemies, StickMelee);
		if(controller_manager_.buttons_down() & SCE_CTRL_R && MyPlayer.has_sword() == true)
			PlayerWeapon.Melee(MyMap.Enemies, SwordMelee);

		// when the right should button is released, the player can fire the gun again
		if(controller_manager_.buttons_released() & SCE_CTRL_R)
			PlayerWeapon.Ranged();

		// propel each bullet across the screen in the direction they were fired
		for(std::vector<GameObject>::iterator bullet = PlayerWeapon.Bullets.begin(); bullet < PlayerWeapon.Bullets.end(); bullet++)
		{
			bullet->FlyForwards();
		}

		// update the enemies
		for(std::vector<GameObject>::iterator enemy = MyMap.Enemies.begin(); enemy < MyMap.Enemies.end(); enemy++)
		{
			// when the enemies are on screen, make them move towards the player
			if(MyCamera.CollisionDetection(*enemy) == true)
				enemy->UpdateEnemy(MyPlayer.position().x, MyPlayer.position().y);

			// when the enemies collide with the player, the player loses health
			if(enemy->CollisionDetection(MyPlayer) == true)
			{
				if(MyPlayer.has_armour() == false)
					MyPlayer.set_health(MyPlayer.health() - EnemyDamage);
				else
					MyPlayer.set_health(MyPlayer.health() - (EnemyDamage * ArmourResistance));
			}
		}

		// loop through all the bullets
		std::vector<GameObject>::iterator bullet = PlayerWeapon.Bullets.begin();
		while(bullet != PlayerWeapon.Bullets.end())
		{
			bool bullet_hit_enemy = false;

			// loop through all the enemies
			std::vector<GameObject>::iterator enemy = MyMap.Enemies.begin();
			while(enemy != MyMap.Enemies.end())
			{
				// when a bullet collides with an enemy, the enemy takes damage
				if(bullet->CollisionDetection(*enemy) == true)
				{
					enemy->set_health(enemy->health() - PlayerWeapon.damage());
					bullet_hit_enemy = true;
				}
				else
					bullet_hit_enemy = false;

				// if an enemy was hit make it flash red
				if(bullet_hit_enemy == true)
					enemy->set_colour(DamagedColour);
				else
					enemy->set_colour(EnemyColour);

				// when an enemy's health reaches zero, delete the enemy
				if(enemy->health() < 0.1)
					MyMap.Enemies.erase(enemy);
				else
					enemy++;
			}
		
			// delete bullets when they leave the screen or collide with an enemy
			if(MyCamera.CollisionDetection(*bullet) == false || bullet_hit_enemy == true)
				PlayerWeapon.Bullets.erase(bullet);
			else
				bullet++;
		}

		// set the impassable door variable to false by default
		impassable_door_ = false;

		// loop through all the doors
		for(std::vector<GameObject>::iterator current_door = MyMap.Doors.begin(); current_door < MyMap.Doors.end(); current_door++)
		{
			// check if the player collides with one of the doors
			if(MyPlayer.CollisionDetection(*current_door) == true)
			{
				// check if the player has enough keys for the door
				if(MyPlayer.number_of_keys() >= current_door->keys_needed())
				{
					// move the player to a position beside the door in the new room
					MyPlayer.MoveTo(current_door->new_door().x + current_door->door_exit(), current_door->new_door().y, MyPlayer.position().z);
					// change the room and create the new rooms' objects
					MyMap.set_current_room(current_door->room_destination());
					MyMap.ChangeRoom();
					// delete all the bullets
					MyMap.ClearRoom(PlayerWeapon.Bullets);
				}
				// if the player needs more keys, set the door impassable variable to true
				else
					impassable_door_ = true;
			}
		}

		// check for when the player picks up items
		if(MyMap.current_room() == FirstKeyRoom && MyPlayer.CollisionDetection(MyMap.Keys[0]) == true && MyPlayer.has_first_key() == false)
		{
			MyPlayer.set_has_first_key(true);
			MyPlayer.set_number_of_keys(MyPlayer.number_of_keys() + 1);
		}
		if(MyMap.current_room() == SecondKeyRoom && MyPlayer.CollisionDetection(MyMap.Keys[1]) == true && MyPlayer.has_second_key() == false)
		{
			MyPlayer.set_has_second_key(true);
			MyPlayer.set_number_of_keys(MyPlayer.number_of_keys() + 1);
		}
		if(MyMap.current_room() == ThirdKeyRoom && MyPlayer.CollisionDetection(MyMap.Keys[2]) == true && MyPlayer.has_third_key() == false)
		{
			MyPlayer.set_has_third_key(true);
			MyPlayer.set_number_of_keys(MyPlayer.number_of_keys() + 1);
		}
		if(MyMap.current_room() == ArmourRoom && MyPlayer.CollisionDetection(MyMap.Armour) == true)
			MyPlayer.set_has_armour(true);
		if(MyMap.current_room() == StickRoom && MyPlayer.CollisionDetection(MyMap.Stick) == true)
			MyPlayer.set_has_stick(true);
		if(MyMap.current_room() == GunUpgradeRoom && MyPlayer.CollisionDetection(MyMap.GunUpgrade) == true)
			MyPlayer.set_has_gun_upgrade(true);
		if(MyMap.current_room() == SwordRoom && MyPlayer.CollisionDetection(MyMap.Sword) == true)
			MyPlayer.set_has_sword(true);
		// the player wins when they pick up the objective
		if(MyMap.current_room() == LastRoom && MyPlayer.CollisionDetection(MyMap.Objective) == true)
			current_game_screen_ = WinScreen;

		// remove item information when the player presses x
		if(controller_manager_.buttons_pressed() & SCE_CTRL_CROSS && MyPlayer.number_of_keys() == OneKey)
			MyMap.Keys[0].set_message_seen(true);
		if(controller_manager_.buttons_pressed() & SCE_CTRL_CROSS && MyPlayer.number_of_keys() == TwoKeys)
			MyMap.Keys[1].set_message_seen(true);
		if(controller_manager_.buttons_pressed() & SCE_CTRL_CROSS && MyPlayer.number_of_keys() == ThreeKeys)
			MyMap.Keys[2].set_message_seen(true);
		if(controller_manager_.buttons_pressed() & SCE_CTRL_CROSS && MyPlayer.has_armour() == true)
			MyMap.Armour.set_message_seen(true);
		if(controller_manager_.buttons_pressed() & SCE_CTRL_CROSS && MyPlayer.has_stick() == true)
			MyMap.Stick.set_message_seen(true);
		if(controller_manager_.buttons_pressed() & SCE_CTRL_CROSS && MyPlayer.has_gun_upgrade() == true)
			MyMap.GunUpgrade.set_message_seen(true);
		if(controller_manager_.buttons_pressed() & SCE_CTRL_CROSS && MyPlayer.has_sword() == true)
			MyMap.Sword.set_message_seen(true);

		// increment the frame count
		frame_count_++;
	}

	bool game_running = true;

	return game_running;
}

void Game::Render()
{
	// set up sprite renderer for drawing
	sprite_renderer_->Begin();

	// change to the current game screen
	ChangeGameScreen();

	// draw the player	
	DrawOnScreen(MyPlayer);

	// keep the gun's centre of rotation on the player's centre
	PlayerWeapon.SetX(MyPlayer.position().x + PlayerWeapon.offset().x);
	PlayerWeapon.SetY(MyPlayer.position().y + PlayerWeapon.offset().y);
	
	// draw the player's gun
	DrawOnScreen(PlayerWeapon);

	// draw each bullet
	for(std::vector<GameObject>::iterator bullet = PlayerWeapon.Bullets.begin(); bullet < PlayerWeapon.Bullets.end(); bullet++)
	{
		DrawOnScreen(*bullet);
	}

	// draw map
	DrawOnScreen(MyMap);

	// draw each enemy
	for(std::vector<GameObject>::iterator enemy = MyMap.Enemies.begin(); enemy < MyMap.Enemies.end(); enemy++)
	{
		DrawOnScreen(*enemy);
	}

	// draw each platform
	for(std::vector<GameObject>::iterator current_platform = MyMap.Platforms.begin(); current_platform < MyMap.Platforms.end(); current_platform++)
	{
		DrawOnScreen(*current_platform);
	}

	// draw each door
	for(std::vector<GameObject>::iterator current_door = MyMap.Doors.begin(); current_door < MyMap.Doors.end(); current_door++)
	{
		DrawOnScreen(*current_door);
	}

	// draw any items in the room until the player picks them up
	if(MyMap.current_room() == FirstKeyRoom && MyPlayer.has_first_key() == false)
		DrawOnScreen(MyMap.Keys[0]);
	if(MyMap.current_room() == SecondKeyRoom && MyPlayer.has_second_key() == false)
		DrawOnScreen(MyMap.Keys[1]);
	if(MyMap.current_room() == ThirdKeyRoom && MyPlayer.has_third_key() == false)
		DrawOnScreen(MyMap.Keys[2]);
	if(MyMap.current_room() == ArmourRoom && MyPlayer.has_armour() == false)
		DrawOnScreen(MyMap.Armour);
	if(MyMap.current_room() == StickRoom && MyPlayer.has_stick() == false)
		DrawOnScreen(MyMap.Stick);
	if(MyMap.current_room() == GunUpgradeRoom && MyPlayer.has_gun_upgrade() == false)
		DrawOnScreen(MyMap.GunUpgrade);
	if(MyMap.current_room() == SwordRoom && MyPlayer.has_sword() == false)
		DrawOnScreen(MyMap.Sword);
	if(MyMap.current_room() == LastRoom)
		DrawOnScreen(MyMap.Objective);

	// display item information
	if(MyPlayer.number_of_keys() == OneKey && MyMap.Keys[0].message_seen() == false)
	{
		font_.RenderText(sprite_renderer_, abfw::Vector3(TextXCoordinate, TextYCoordinate, 0), MessageSize, TextColour, abfw::TJ_LEFT,
					 "Obtained first key.  ");
		font_.RenderText(sprite_renderer_, abfw::Vector3(TextXCoordinate, TextYCoordinate + TextOffset, 0), MessageSize, TextColour, abfw::TJ_LEFT,
					 "Press X.  ");
	}
	if(MyPlayer.number_of_keys() == TwoKeys && MyMap.Keys[1].message_seen() == false)
	{
		font_.RenderText(sprite_renderer_, abfw::Vector3(TextXCoordinate, TextYCoordinate, 0), MessageSize, TextColour, abfw::TJ_LEFT,
					 "Obtained second key.  ");
		font_.RenderText(sprite_renderer_, abfw::Vector3(TextXCoordinate, TextYCoordinate + TextOffset, 0), MessageSize, TextColour, abfw::TJ_LEFT,
					 "Press X.  ");
	}
	if(MyPlayer.number_of_keys() == ThreeKeys && MyMap.Keys[2].message_seen() == false)
	{
		font_.RenderText(sprite_renderer_, abfw::Vector3(TextXCoordinate, TextYCoordinate, 0), MessageSize, TextColour, abfw::TJ_LEFT,
					 "Obtained third key.  ");
		font_.RenderText(sprite_renderer_, abfw::Vector3(TextXCoordinate, TextYCoordinate + TextOffset, 0), MessageSize, TextColour, abfw::TJ_LEFT,
					 "Press X.  ");
	}
	if(MyPlayer.has_armour() == true && MyMap.Armour.message_seen() == false)
	{
		font_.RenderText(sprite_renderer_, abfw::Vector3(TextXCoordinate, TextYCoordinate, 0), MessageSize, TextColour, abfw::TJ_LEFT,
					 "Obtained armour.  ");
		font_.RenderText(sprite_renderer_, abfw::Vector3(TextXCoordinate, TextYCoordinate + TextOffset, 0), MessageSize, TextColour, abfw::TJ_LEFT,
					 "Player takes less damage from enemies.  ");
		font_.RenderText(sprite_renderer_, abfw::Vector3(TextXCoordinate, TextYCoordinate + (TextOffset * 2), 0), MessageSize, TextColour, abfw::TJ_LEFT,
					 "Press X.  ");
	}
	if(MyPlayer.has_stick() == true && MyMap.Stick.message_seen() == false)
	{
		font_.RenderText(sprite_renderer_, abfw::Vector3(TextXCoordinate, TextYCoordinate, 0), MessageSize, TextColour, abfw::TJ_LEFT,
					 "Obtained stick. ");
		font_.RenderText(sprite_renderer_, abfw::Vector3(TextXCoordinate, TextYCoordinate + TextOffset, 0), MessageSize, TextColour, abfw::TJ_LEFT,
					 "The stick knocks back enemies.  Press and hold R to use.  ");
		font_.RenderText(sprite_renderer_, abfw::Vector3(TextXCoordinate, TextYCoordinate + (TextOffset * 2), 0), MessageSize, TextColour, abfw::TJ_LEFT,
					 "Press X.  ");
	}
	if(MyPlayer.has_gun_upgrade() == true && MyMap.GunUpgrade.message_seen() == false)
	{
		font_.RenderText(sprite_renderer_, abfw::Vector3(TextXCoordinate, TextYCoordinate, 0), MessageSize, TextColour, abfw::TJ_LEFT,
					 "Obtained gun upgrade.  ");
		font_.RenderText(sprite_renderer_, abfw::Vector3(TextXCoordinate, TextYCoordinate + TextOffset, 0), MessageSize, TextColour, abfw::TJ_LEFT,
					 "The gun now has an increased rate of fire.  ");
		font_.RenderText(sprite_renderer_, abfw::Vector3(TextXCoordinate, TextYCoordinate + (TextOffset * 2), 0), MessageSize, TextColour, abfw::TJ_LEFT,
					 "Press X.  ");
	}
	if(MyPlayer.has_sword() == true && MyMap.Sword.message_seen() == false)
	{
		font_.RenderText(sprite_renderer_, abfw::Vector3(TextXCoordinate, TextYCoordinate, 0), MessageSize, TextColour, abfw::TJ_LEFT,
					 "Obtained sword. The sword replaces the stick and");
		font_.RenderText(sprite_renderer_, abfw::Vector3(TextXCoordinate, TextYCoordinate + TextOffset, 0), MessageSize, TextColour, abfw::TJ_LEFT,
					 "inflicts heavy damage on enemies.  ");
		font_.RenderText(sprite_renderer_, abfw::Vector3(TextXCoordinate, TextYCoordinate + (TextOffset * 2), 0), MessageSize, TextColour, abfw::TJ_LEFT,
					 "Press X.  ");
	}

	// let the player know if a door is impassable
	if(impassable_door_ == true)
		font_.RenderText(sprite_renderer_, abfw::Vector3(TextXCoordinate, TextYCoordinate, 0), MessageSize, TextColour, abfw::TJ_LEFT,
					 "You need more keys to get passed this door.  ");
	
	// display the player's health
	font_.RenderText(sprite_renderer_, abfw::Vector3(HealthDisplayXCoordinate, HealthDisplayYCoordinate, 0.f), HealthDisplaySize, HealthDisplayColour,
					 abfw::TJ_LEFT, "Health = %.1f", MyPlayer.health());

	// tell sprite renderer that all sprites have been drawn
	sprite_renderer_->End();
}

void Game::DrawOnScreen(GameObject& Object)
{
	// if the object is on screen
	if(MyCamera.CollisionDetection(Object) == true)
	{
		// take the objects current position as world coordinates
		abfw::Vector3 world_coordinates = Object.position();

		// convert its position to screen coordinates
		abfw::Vector3 screen_coordinates(Object.position().x - (MyCamera.position().x - MyCamera.HalfSize(MyCamera.width())),
		Object.position().y - (MyCamera.position().y - MyCamera.HalfSize(MyCamera.height())), Object.position().z);
		
		// move the object to the screen coordinates
		Object.MoveTo(screen_coordinates.x, screen_coordinates.y, screen_coordinates.z);

		// draw the object
		sprite_renderer_->DrawSprite(Object);

		// move the object back to its world coordinates
		Object.MoveTo(world_coordinates.x, world_coordinates.y, world_coordinates.z);
	}

}

void Game::ChangeGameScreen()
{
	// change the game state based on the game state variable
	switch(current_game_screen_)
	{
		case StartScreen:
			// display the text, "PRESS START"
			font_.RenderText(sprite_renderer_, abfw::Vector3(TextXCoordinate, TextYCoordinate, 0), ScreenTextSize, TextColour, abfw::TJ_LEFT,
							 "PRESS START");
			// reset player properties
			MyPlayer.MoveTo(MyMap.position().x, MyMap.position().y, 0);
			MyPlayer.set_colour(PlayerColour);
			MyPlayer.set_health(PlayerHealth);
			MyPlayer.set_number_of_keys(NoKeys);
			MyPlayer.set_has_first_key(false);
			MyPlayer.set_has_second_key(false);
			MyPlayer.set_has_third_key(false);
			MyPlayer.set_has_armour(false);
			MyPlayer.set_has_stick(false);
			MyPlayer.set_has_gun_upgrade(false);
			MyPlayer.set_has_sword(false);
			// reset the player's weapon
			PlayerWeapon.MoveTo(MyPlayer.position().x + GunOffset, MyPlayer.position().y, 0);
			PlayerWeapon.Ranged();
			// centre the camera on the player's position
			MyCamera.MoveTo(MyPlayer.position().x, MyPlayer.position().y, 0);
			// start the game in the main room
			MyMap.set_current_room(MainRoom);
			MyMap.ChangeRoom();
			MyMap.ClearRoom(PlayerWeapon.Bullets);
			// reset the item information messages
			for(int key_message = 0; key_message < 3; key_message++)
			{
				MyMap.Keys[key_message].set_message_seen(false);
			}
			MyMap.Armour.set_message_seen(false);
			MyMap.Stick.set_message_seen(false);
			MyMap.GunUpgrade.set_message_seen(false);
			MyMap.Sword.set_message_seen(false);
			// start the game when the player presses start
			if(controller_manager_.buttons_pressed() & SCE_CTRL_START)
			{
				current_game_screen_ = GameScreen;
			}
			break;
		case GameOverScreen:
			// display the text, "GAME OVER"
			font_.RenderText(sprite_renderer_, abfw::Vector3(TextXCoordinate, TextYCoordinate, 0), ScreenTextSize, TextColour, abfw::TJ_LEFT,
							 "GAME OVER");
			// reset the game when the player preses start
			if(controller_manager_.buttons_pressed() & SCE_CTRL_START)
				current_game_screen_ = StartScreen;
			break;
		case PauseScreen:
			// display the text, "PRESS START"
			font_.RenderText(sprite_renderer_, abfw::Vector3(TextXCoordinate, TextYCoordinate, 0), ScreenTextSize, TextColour, abfw::TJ_LEFT,
							 "PAUSED");
			// resume the game when the player presses start
			if(controller_manager_.buttons_pressed() & SCE_CTRL_START)
				current_game_screen_ = GameScreen;
			break;
		case GameScreen:
			// pause the game when the player presses start
			if(controller_manager_.buttons_pressed() & SCE_CTRL_START)
				current_game_screen_ = PauseScreen;
			// end the game when the players health reaches zero
			if(MyPlayer.health() < 0.1f)
			{
				MyPlayer.set_health(0);
				current_game_screen_ = GameOverScreen;
			}
			break;
		case WinScreen:
			// display the text, "YOU WIN"
			font_.RenderText(sprite_renderer_, abfw::Vector3(TextXCoordinate, TextYCoordinate, 0), ScreenTextSize, TextColour, abfw::TJ_LEFT,
							 "YOU WIN");
			// reset the game when the player preses start
			if(controller_manager_.buttons_pressed() & SCE_CTRL_START)
				current_game_screen_ = StartScreen;
			break;
	}
}
 